/*
    Copyright 2011 Robert Baruch (robert.c.baruch@gmail.com).
    Support at diybookscanner.org.

    This file is part of QR Perspective Correction.

    QR Perspective Correction is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    QR Perspective Correction is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with QR Perspective Correction.  If not, see <http://www.gnu.org/licenses/>.

 */
package qrpc;

import com.google.zxing.*;
import com.google.zxing.client.j2se.BufferedImageLuminanceSource;
import com.google.zxing.common.BitMatrix;
import com.google.zxing.common.GlobalHistogramBinarizer;
import com.google.zxing.common.HybridBinarizer;
import com.google.zxing.multi.qrcode.QRCodeMultiReader;
import com.sun.imageio.plugins.jpeg.JPEGImageWriter;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Iterator;
import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import org.apache.commons.math.linear.*;
import org.apache.log4j.*;
import org.w3c.dom.Element;

/**
 *
 * @author robertbaruch
 */
public class Main
{
    public static Logger log = Logger.getLogger(Main.class);
    static double[] m = null; // dekeystone parameters
    static int dpi = 72; // default

    static
    {
        Logger.getRootLogger().addAppender(new
                ConsoleAppender(new
                EnhancedPatternLayout("%d [%-5p] %m%n")));
    }

    private static class PartialCodesFoundException extends Exception
    {

    }

    private static void warp(double[] dest, double[] src, double[] m)
    {
        double w = m[6]*dest[0] + m[7]*dest[1] + 1;
        src[0] = (m[0]*dest[0] + m[1]*dest[1] + m[2])/w;
        src[1] = (m[3]*dest[0] + m[4]*dest[1] + m[5])/w;
    }

    private static void writeBinaryImage(BitMatrix matrix, File out) throws IOException
    {
        int w = matrix.width;
        int h = matrix.height;
        BufferedImage image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        for (int y=0; y<h; y++)
            for (int x=0; x<w; x++)
                image.setRGB(x, y, matrix.get(x, y) ? 0 : 0xFFFFFF);
        writeImage(image, out);

    }

    private static void writeImage(BufferedImage image, File out) throws IOException
    {
            // Image writer
            ImageWriter imageWriter = null;
            Iterator<ImageWriter> iterator = ImageIO.getImageWritersByFormatName("jpeg");
            while (iterator.hasNext())
            {
                imageWriter = iterator.next();
                if (imageWriter instanceof JPEGImageWriter)
                    break;
                imageWriter = null;
            }

            if (imageWriter == null)
            {
                log.fatal("Couldn't find an image writer for JPEG images. Giving up.");
                System.exit(0);
            }

            ImageOutputStream ios = ImageIO.createImageOutputStream(out);
            imageWriter.setOutput(ios);

            // Compression
            ImageWriteParam jpegParams = imageWriter.getDefaultWriteParam();
            jpegParams.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
            jpegParams.setCompressionQuality(1.0f);

            // Metadata (dpi)
            IIOMetadata data = imageWriter.getDefaultImageMetadata(new ImageTypeSpecifier(image), jpegParams);
            Element tree = (Element)data.getAsTree("javax_imageio_jpeg_image_1.0");
            Element jfif = (Element)tree.getElementsByTagName("app0JFIF").item(0);
            jfif.setAttribute("Xdensity", Integer.toString(dpi));
            jfif.setAttribute("Ydensity", Integer.toString(dpi));
            jfif.setAttribute("resUnits", "1"); // density is dots per inch
            data.setFromTree("javax_imageio_jpeg_image_1.0", tree);

            // Write and clean up
            imageWriter.write(null, new IIOImage(image, null, data), jpegParams);
            ios.close();
            imageWriter.dispose();

    }

    private static Result[] locateQR(Binarizer binarizer)
            throws PartialCodesFoundException
    {
        Result TL = null;
        Result TR = null;
        Result BL = null;
        Result BR = null;

        try
        {
            QRCodeMultiReader detector = new QRCodeMultiReader();
            Hashtable hints = new Hashtable();
            hints.put(DecodeHintType.TRY_HARDER, Boolean.TRUE);
            Result[] results = detector.decodeMultiple(new BinaryBitmap(binarizer), hints);
            log.debug("Number of QR codes found: " + results.length);

            if (results.length == 0)
                return null;


            for (int i=0; i<results.length; i++)
            {
                log.debug("QR code '" + results[i].getText() + "':");
                for (int j=0; j<results[i].getResultPoints().length; j++)
                {
                    log.debug("  " + results[i].getResultPoints()[j].getX() + "," +
                            results[i].getResultPoints()[j].getY());
                }
                if (results[i].getText().equals("DIYBookscannerTL"))
                    TL = results[i];
                else if (results[i].getText().equals("DIYBookscannerTR"))
                    TR = results[i];
                else if (results[i].getText().equals("DIYBookscannerBL"))
                    BL = results[i];
                else if (results[i].getText().equals("DIYBookscannerBR"))
                    BR = results[i];
            }

        } catch (NotFoundException ex)
        {
            log.info("No QR codes found in image");
            return null;
        }

        if (BR == null || TR == null || BL == null || TL == null)
        {
            log.info("Located SOME of the expected QR codes, but not ALL.");
            throw new PartialCodesFoundException();
        }

        return new Result[] { TL, TR, BL, BR };
    }

    private static double[] findQR(BufferedImage image) throws Exception
    {
        log.debug("Create luminance source");
        BufferedImageLuminanceSource luminanceSource =
                new BufferedImageLuminanceSource(image);
        log.debug("Create binarizer");
        Binarizer binarizer = new GlobalHistogramBinarizer(luminanceSource);
        //writeBinaryImage(binarizer.getBlackMatrix(), new File("binary.jpg"));
        log.debug("Look for QR codes");

        Result TL = null;
        Result TR = null;
        Result BL = null;
        Result BR = null;

        try
        {
            Result[] results = locateQR(binarizer);

            if (results == null)
                return null;

            TL = results[0];
            TR = results[1];
            BL = results[2];
            BR = results[3];
        }
        catch (PartialCodesFoundException ex)
        {
            // try again, but with the HybridBinarizer

            log.warn("Didn't find all expected QR codes: trying alternate binarization algorithm");
            binarizer = new HybridBinarizer(luminanceSource);
            //writeBinaryImage(binarizer.getBlackMatrix(), new File("binary2.jpg"));
            try
            {
                Result[] results = locateQR(binarizer);

                if (results == null)
                    return null;

                TL = results[0];
                TR = results[1];
                BL = results[2];
                BR = results[3];
            }
            catch (PartialCodesFoundException ex2)
            {
                log.error("!!! Still could not locate all QR codes.");
                log.error("!!! This registration image will be ignored.");
                throw new PartialCodesFoundException();
            }
        }


        double s_TRx = TR.getResultPoints()[0].getX();
        double s_TRy = TR.getResultPoints()[0].getY();
        double s_TLx = TL.getResultPoints()[0].getX();
        double s_TLy = TL.getResultPoints()[0].getY();
        double s_BRx = BR.getResultPoints()[0].getX();
        double s_BRy = BR.getResultPoints()[0].getY();
        double s_BLx = BL.getResultPoints()[0].getX();
        double s_BLy = BL.getResultPoints()[0].getY();

        double side = (float)Math.sqrt((s_TRx-s_TLx)*(s_TRx-s_TLx) + (s_TRy-s_TLy)*(s_TRy-s_TLy));
        dpi = (int)(side/3);

        log.info("Estimated dpi: " + dpi);

        double d_TLx = s_TLx;
        double d_TLy = s_TLy;
        double d_TRx = d_TLx + side;
        double d_TRy = d_TLy;
        double d_BLx = d_TLx;
        double d_BLy = d_TLy + side;
        double d_BRx = d_TLx + side;
        double d_BRy = d_TLy + side;

        double[][] dmatrix = // [row][column]
        {
            { d_TLx, d_TLy, 1, 0, 0,         0, -d_TLx*s_TLx, -d_TLy*s_TLx },
            { d_TRx, d_TRy, 1, 0, 0,         0, -d_TRx*s_TRx, -d_TRy*s_TRx },
            { d_BLx, d_BLy, 1, 0, 0,         0, -d_BLx*s_BLx, -d_BLy*s_BLx },
            { d_BRx, d_BRy, 1, 0, 0,         0, -d_BRx*s_BRx, -d_BRy*s_BRx },
            { 0, 0,         0, d_TLx, d_TLy, 1, -d_TLx*s_TLy, -d_TLy*s_TLy },
            { 0, 0,         0, d_TRx, d_TRy, 1, -d_TRx*s_TRy, -d_TRy*s_TRy },
            { 0, 0,         0, d_BLx, d_BLy, 1, -d_BLx*s_BLy, -d_BLy*s_BLy },
            { 0, 0,         0, d_BRx, d_BRy, 1, -d_BRx*s_BRy, -d_BRy*s_BRy }
        };

        Array2DRowRealMatrix matrix = new Array2DRowRealMatrix(dmatrix);
        LUDecompositionImpl decomposition = new LUDecompositionImpl(matrix);
        DecompositionSolver solver = decomposition.getSolver();
        double[] mm = solver.solve(new double[]
                { s_TLx, s_TRx, s_BLx, s_BRx, s_TLy, s_TRy, s_BLy, s_BRy });

        log.debug("m00 " + mm[0]);
        log.debug("m01 " + mm[1]);
        log.debug("m02 " + mm[2]);
        log.debug("m10 " + mm[3]);
        log.debug("m11 " + mm[4]);
        log.debug("m12 " + mm[5]);
        log.debug("m20 " + mm[6]);
        log.debug("m21 " + mm[7]);

        double[] dest = new double[2];
        double[] src = new double[2];

        warp(new double[] { d_TLx, d_TLy }, src, mm);
        log.debug(" TL (" + d_TLx + "," + d_TLy + ") -> (" + src[0] + "," + src[1] + ")");
        warp(new double[] { d_TRx, d_TRy }, src, mm);
        log.debug(" TR (" + d_TRx + "," + d_TRy + ") -> (" + src[0] + "," + src[1] + ")");
        warp(new double[] { d_BLx, d_BLy }, src, mm);
        log.debug(" BL (" + d_BLx + "," + d_BLy + ") -> (" + src[0] + "," + src[1] + ")");
        warp(new double[] { d_BRx, d_BRy }, src, mm);
        log.debug(" BR (" + d_BRx + "," + d_BRy + ") -> (" + src[0] + "," + src[1] + ")");

        return mm;
    }

    public static void main(String[] args) throws Exception
    {
        System.out.println("qrpc (QR Perspective Correction) v 1.6 (20110213A), by Robert Baruch (robert.c.baruch@gmail.com)");
        
        if (args.length != 1)
        {
            System.err.println("Usage: java -jar qrpc.jar <directory with image files>");
            return;
        }

        File dir = new File(args[0]);
        if (!dir.isDirectory())
        {
            System.err.println("'" + args[0] + "' is not a directory.");
            System.err.println("Usage: java -jar qrpc.jar <directory with image files>");
            return;
        }
        
        File[] files = dir.listFiles(new FileFilter() {
            public boolean accept(File file)
            {
                return file.isFile() &&
                        (file.getName().toUpperCase().endsWith(".JPG") ||
                         file.getName().toUpperCase().endsWith(".JPEG")) &&
                         !file.getName().startsWith("dekey ") &&
                         !file.getName().startsWith("qr ");
            }

        });

        Arrays.sort(files, new Comparator<File>() {
            public int compare(File t, File t1)
            {
                return t.getName().compareTo(t1.getName());
            }
         });

        log.info("Found " + files.length + " jpeg images to process");

        for (File file : files)
        {
            log.debug("Loading image " + file.getName());
            BufferedImage image = ImageIO.read(file);

            double[] mm;
            
            try
            {
                mm = findQR(image);
            }
            catch (PartialCodesFoundException ex)
            {
                File newfile = new File(dir, "problem " + file.getName() );
                writeImage(image, newfile);
                log.warn("Partial registration image: saved as file 'problem " + file.getName() + "'");
                continue;
            }

            if (mm == null && m == null)
            {
                log.warn("No initial QR codes found for dekeystoning. Leaving file alone.");
                continue;
            }

            log.info("Dekeystoning image...");

            PerspectiveOp op = new PerspectiveOp(mm == null ? m : mm, PerspectiveOp.Type.NEAREST_NEIGHBOR, null);

            BufferedImage image2 = op.filter(image, null);
            File newfile = new File(dir, (mm == null ? "dekey " : "qr ") + file.getName() );

            writeImage(image2, newfile);

            log.info("Image dekeystoned, saved to file " + newfile.getName());

            if (mm != null) // new parameters
            {
                m = mm;
            }

        }
   }

}
