/*
    Copyright 2011 Robert Baruch (robert.c.baruch@gmail.com).
    Support at diybookscanner.org.

    This file is part of QR Perspective Correction.

    QR Perspective Correction is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    QR Perspective Correction is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with QR Perspective Correction.  If not, see <http://www.gnu.org/licenses/>.

 */
package qrpc;

import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.Raster;
import java.awt.image.RasterFormatException;
import java.awt.image.WritableRaster;
import java.util.Arrays;

/**
 *
 * @author robertbaruch
 */
public class PerspectiveOp implements BufferedImageOp
{
    public static enum Type
    {
        BICUBIC, BILINEAR, NEAREST_NEIGHBOR
    }

    double[] m;
    Type interpolationType;
    RenderingHints hints;

    public PerspectiveOp(double[] xform, Type interpolationType, RenderingHints hints)
    {
        if (xform == null || xform.length != 8)
            throw new IllegalArgumentException("xform must be an array of 8 doubles");
        m = xform;
        this.interpolationType = interpolationType;
        this.hints = hints;
    }

    private void getBilinearPixel(int x0, int y0, int x1, int y1, int x2, int y2, int x3, int y3,
            Rectangle2D srcBounds, Raster srcRaster, int[] sample)
    {
        if (srcBounds.contains(x0, y0))
        {
            srcRaster.getPixel(x0, y0, sample);
            return;
        }

        if (srcBounds.contains(x1, y1))
        {
            srcRaster.getPixel(x1, y1, sample);
            return;
        }

        if (srcBounds.contains(x2, y2))
        {
            srcRaster.getPixel(x2, y2, sample);
            return;
        }

        srcRaster.getPixel(x3, y3, sample);

    }

    public BufferedImage filter(BufferedImage src, BufferedImage dst)
    {
        if (src == null)
            throw new NullPointerException("src image must not be null");
        if (dst != null)
            throw new IllegalArgumentException("dst image must be null");

        ColorModel srcCM = src.getColorModel();
        ColorModel dstCM = srcCM;
        BufferedImage origDst = dst;

        dst = createCompatibleDestImage(src, null);
        int w = dst.getWidth();
        int h = dst.getHeight();
        Point2D srcPt = new Point2D.Double();
        Point2D dstPt = new Point2D.Double();
        Raster srcRaster = src.getData();
        WritableRaster dstRaster = dst.getRaster();
        Rectangle2D srcBounds = srcRaster.getBounds();

        int[] ZERO = new int[srcRaster.getNumBands()];
        Arrays.fill(ZERO, 0);

        int[] sample = new int[srcRaster.getNumBands()];

        // For interpolation

        int[] sample1 = new int[srcRaster.getNumBands()];
        int[] sample2 = new int[srcRaster.getNumBands()];
        int[] sample3 = new int[srcRaster.getNumBands()];
        int[] sample4 = new int[srcRaster.getNumBands()];

        for (int x=0; x<w; x++)
            for (int y=0; y<h; y++)
            {
                dstPt.setLocation(x, y);
                getPoint2D(dstPt, srcPt);
                if (interpolationType == Type.NEAREST_NEIGHBOR)
                {
                    int sx = (int)Math.round(srcPt.getX());
                    int sy = (int)Math.round(srcPt.getY());
                    if (!srcBounds.contains(sx, sy))
                        dstRaster.setPixel(x, y, ZERO);
                    else
                        dstRaster.setPixel(x, y, srcRaster.getPixel(sx, sy, sample));
                }
                else if (interpolationType == Type.BILINEAR)
                {
                    int xfloor = (int)Math.floor(srcPt.getX());
                    int xceil = (int)Math.ceil(srcPt.getX());
                    int yfloor = (int)Math.floor(srcPt.getY());
                    int yceil = (int)Math.ceil(srcPt.getY());

                    if (!srcBounds.contains(xfloor, yfloor) &&
                            !srcBounds.contains(xceil, yceil))
                    {
                        dstRaster.setPixel(x, y, ZERO);
                        continue;
                    }

                    // if not all four points are within the image,
                    // then we have to fall back to other points.

                    if (srcBounds.contains(xfloor, yfloor) &&
                            srcBounds.contains(xceil, yceil))
                    {
                        srcRaster.getPixel(xfloor, yfloor, sample1);
                        srcRaster.getPixel(xfloor, yceil, sample2);
                        srcRaster.getPixel(xceil, yfloor, sample3);
                        srcRaster.getPixel(xceil, yceil, sample4);
                    }
                    else
                    {
                        getBilinearPixel(xfloor, yfloor,
                                xfloor, yceil,
                                xceil, yfloor,
                                xceil, yceil,
                                srcBounds, srcRaster, sample1);
                        getBilinearPixel(xfloor, yceil,
                                xfloor, yfloor,
                                xceil, yceil,
                                xceil, yfloor,
                                srcBounds, srcRaster, sample2);
                        getBilinearPixel(xceil, yfloor,
                                xceil, yceil,
                                xfloor, yfloor,
                                xfloor, yceil,
                                srcBounds, srcRaster, sample3);
                        getBilinearPixel(xceil, yceil,
                                xceil, yfloor,
                                xfloor, yceil,
                                xfloor, yfloor,
                                srcBounds, srcRaster, sample3);
                    }

                    double xfrac = srcPt.getX() - xfloor;
                    double yfrac = srcPt.getY() - yfloor;

                    // interpolate along x axis

                    for (int i=0; i<sample.length; i++)
                    {
                        sample1[i] += (sample3[i] - sample1[i])*xfrac;
                        sample2[i] += (sample4[i] - sample2[i])*xfrac;
                    }

                    // interpolate along y axis

                    for (int i=0; i<sample.length; i++)
                        sample1[i] += (sample2[i] - sample1[i])*yfrac;

                    dstRaster.setPixel(x, y, sample1);
                }
            }
        
        return dst;

    }

    public Rectangle2D getBounds2D(BufferedImage src)
    {
        return new Rectangle2D.Double(0, 0, src.getWidth(), src.getHeight());
    }

    public BufferedImage createCompatibleDestImage(BufferedImage src, ColorModel destCM)
    {
        BufferedImage image;
        Rectangle r = getBounds2D(src).getBounds();

        // If r.x (or r.y) is < 0, then we want to only create an image
        // that is in the positive range.
        // If r.x (or r.y) is > 0, then we need to create an image that
        // includes the translation.
        int w = r.x + r.width;
        int h = r.y + r.height;
        if (w <= 0) {
            throw new RasterFormatException("Transformed width ("+w+
                                            ") is less than or equal to 0.");
        }
        if (h <= 0) {
            throw new RasterFormatException("Transformed height ("+h+
                                            ") is less than or equal to 0.");
        }

        if (destCM == null) {
            ColorModel cm = src.getColorModel();
            if (interpolationType != Type.NEAREST_NEIGHBOR &&
                (cm instanceof IndexColorModel ||
                 cm.getTransparency() == Transparency.OPAQUE))
            {
                image = new BufferedImage(w, h,
                                          BufferedImage.TYPE_INT_ARGB);
            }
            else {
                image = new BufferedImage(cm,
                          src.getRaster().createCompatibleWritableRaster(w,h),
                          cm.isAlphaPremultiplied(), null);
            }
        }
        else {
            image = new BufferedImage(destCM,
                                    destCM.createCompatibleWritableRaster(w,h),
                                    destCM.isAlphaPremultiplied(), null);
        }

        return image;
    }

    /**
     * Transforms the given destination point into a source point, returning
     * the source point.
     *
     * @param dest The destination point to transform.
     * @param src The source point to return. May be null.
     * @return The source point transformed from the destination point.
     */
    public Point2D getPoint2D(Point2D dest, Point2D src)
    {
        if (src == null)
            src = new Point2D.Double();
        double w = m[6]*dest.getX() + m[7]*dest.getY() + 1;
        src.setLocation( (m[0]*dest.getX() + m[1]*dest.getY() + m[2])/w ,
                         (m[3]*dest.getX() + m[4]*dest.getY() + m[5])/w );
        return src;
    }

    public RenderingHints getRenderingHints()
    {
        return hints;
    }

}
